@@ -13,7 +13,7 @@
 ##############################################################################
 """RML Attribute Implementation
 
-$Id: attr.py 121449 2011-04-19 16:00:38Z menesis $
+$Id: attr.py 128842 2012-12-21 05:38:51Z srichter $
 """
 __docformat__ = "reStructuredText"
 import cStringIO
@@ -37,12 +37,12 @@ MISSING = object()
 logger = logging.getLogger("z3c.rml")
 
 
-def getFileInfo(attr):
-    root = attr.context
+def getFileInfo(directive):
+    root = directive
     while root.parent:
         root = root.parent
     return '(file %s, line %i)' % (
-        root.filename, attr.context.element.sourceline)
+        root.filename, directive.element.sourceline)
 
 
 def getManager(context, interface=None):
@@ -87,7 +87,7 @@ class RMLAttribute(zope.schema.Field):
             name = self.deprecatedName
             logger.warn(
                 u'Deprecated attribute "%s": %s %s' % (
-                name, self.deprecatedReason, getFileInfo(self)))
+                name, self.deprecatedReason, getFileInfo(self.context)))
         else:
             name = self.__name__
         # Extract the value.
@@ -102,14 +102,16 @@ class RMLAttribute(zope.schema.Field):
 
 class BaseChoice(RMLAttribute):
     choices = {}
+    doLower = True
 
     def fromUnicode(self, value):
-        value = value.lower()
+        if self.doLower:
+            value = value.lower()
         if value in self.choices:
             return self.choices[value]
         raise ValueError(
             '%r not a valid value for attribute "%s". %s' % (
-            value, self.__name__, getFileInfo(self)))
+            value, self.__name__, getFileInfo(self.context)))
 
 
 class Combination(RMLAttribute):
@@ -127,7 +129,8 @@ class Combination(RMLAttribute):
             except ValueError:
                 pass
         raise ValueError(
-            '"%s" is not a valid value. %s' %(value, getFileInfo(self)))
+            '"%s" is not a valid value. %s' %(
+                value, getFileInfo(self.context)))
 
 
 class String(RMLAttribute, zope.schema.Bytes):
@@ -185,18 +188,46 @@ class Sequence(RMLAttribute, zope.schema._field.AbstractCollection):
             (self.max_length is not None and len(result) > self.max_length)):
             raise ValueError(
                 'Length of sequence must be at least %s and at most %i. %s' % (
-                self.min_length, self.max_length, getFileInfo(self)))
+                self.min_length, self.max_length,
+                getFileInfo(self.context)))
         return result
 
 
+class IntegerSequence(Sequence):
+    """A sequence of integers."""
+
+    def fromUnicode(self, ustr):
+        ustr = ustr.strip()
+        pieces = self.splitre.split(ustr)
+        numbers = set([])
+        for piece in pieces:
+            # Ignore empty pieces.
+            if not piece:
+                continue
+            # The piece is a range.
+            if '-' in piece:
+                start, end = piece.split('-')
+                # Make range lower and upper bound inclusive.
+                numbers.update(range(int(start), int(end)+1))
+                continue
+            # The piece is just a number
+            numbers.add(int(piece))
+        return list(numbers)
+
 class Choice(BaseChoice):
     """A choice of several values. The values are always case-insensitive."""
 
-    def __init__(self, choices=None, *args, **kw):
+    def __init__(self, choices=None, doLower=True, *args, **kw):
         super(Choice, self).__init__(*args, **kw)
         if isinstance(choices, (tuple, list)):
-            choices = dict([(val.lower(), val) for val in choices])
+            choices = dict(
+                [(val.lower() if doLower else val, val) for val in choices])
+        else:
+            choices = dict(
+                [(key.lower() if doLower else key, val)
+                 for key, val in choices.items()])
         self.choices = choices
+        self.doLower = doLower
 
 
 class Boolean(BaseChoice):
@@ -220,7 +251,7 @@ class BooleanWithDefault(Boolean):
 class Measurement(RMLAttribute):
     '''This field represents a length value.
 
-    The units "in" (inch), "cm", and "mm" are allowed. If no units are
+    The units "in" (inch), "cm", "mm", and "pt" are allowed. If no units are
     specified, the value is given in points/pixels.
     '''
     def __init__(self, allowPercentage=False, allowStar=False, *args, **kw):
@@ -232,6 +263,7 @@ class Measurement(RMLAttribute):
 	(re.compile('^(-?[0-9\.]+)\s*in$'), reportlab.lib.units.inch),
 	(re.compile('^(-?[0-9\.]+)\s*cm$'), reportlab.lib.units.cm),
 	(re.compile('^(-?[0-9\.]+)\s*mm$'), reportlab.lib.units.mm),
+	(re.compile('^(-?[0-9\.]+)\s*pt$'), 1),
 	(re.compile('^(-?[0-9\.]+)\s*$'), 1)
         ]
 
@@ -251,7 +283,7 @@ class Measurement(RMLAttribute):
                 return unit[1]*float(res.group(1))
         raise ValueError(
             'The value %r is not a valid measurement. %s' % (
-            value, getFileInfo(self)))
+            value, getFileInfo(self.context)))
 
 
 class File(Text):
@@ -277,7 +309,7 @@ class File(Text):
             if result is None:
                 raise ValueError(
                     'The package-path-pair you specified was incorrect. %s' %(
-                    getFileInfo(self)))
+                    getFileInfo(self.context)))
             modulepath, path = result.groups()
             module = __import__(modulepath, {}, {}, (modulepath))
             value = os.path.join(os.path.dirname(module.__file__), path)
@@ -325,9 +357,13 @@ class Color(RMLAttribute):
         self.acceptNone = acceptNone
 
     def fromUnicode(self, value):
-        if self.acceptNone and value == 'None':
+        if self.acceptNone and value.lower() == 'none':
             return None
         manager = getManager(self.context)
+
+        if value.startswith('rml:'):
+            value = manager.names[value[4:]]
+
         if value in manager.colors:
             return manager.colors[value]
         try:
@@ -336,7 +372,20 @@ class Color(RMLAttribute):
         except:
             raise ValueError(
                 'The color specification "%s" is not valid. %s' % (
-                value, getFileInfo(self)))
+                value, getFileInfo(self.context)))
+
+def _getStyle(context, value):
+    manager = getManager(context)
+    for styles in (manager.styles,
+                   reportlab.lib.styles.getSampleStyleSheet().byName):
+        if value in styles:
+            return styles[value]
+        elif 'style.' + value in styles:
+            return styles['style.' + value]
+        elif value.startswith('style.') and value[6:] in styles:
+            return styles[value[6:]]
+    raise ValueError('Style %r could not be found. %s' % (
+        value, getFileInfo(context)))
 
 class Style(String):
     """Requires a valid style to be entered.
@@ -347,17 +396,23 @@ class Style(String):
     default = reportlab.lib.styles.getSampleStyleSheet().byName['Normal']
 
     def fromUnicode(self, value):
-        manager = getManager(self.context)
-        for styles in (manager.styles,
-                       reportlab.lib.styles.getSampleStyleSheet().byName):
-            if value in styles:
-                return styles[value]
-            elif 'style.' + value in styles:
-                return styles['style.' + value]
-            elif value.startswith('style.') and value[6:] in styles:
-                return styles[value[6:]]
-        raise ValueError('Style %r could not be found. %s' % (
-            value, getFileInfo(self)))
+        return _getStyle(self.context, value)
+
+
+class Padding(Sequence):
+    """This attribute is specific for padding and will produce the proper
+    length of the padding sequence."""
+
+    def __init__(self, *args, **kw):
+        kw.update(dict(value_type=Integer(), min_length=1, max_length=4))
+        super(Padding, self).__init__(*args, **kw)
+
+    def fromUnicode(self, value):
+        seq = super(Padding, self).fromUnicode(value)
+        # pdfgen does not like a single paddign value.
+        if len(seq) == 1:
+            seq.append(seq[0])
+        return seq
 
 
 class Symbol(Text):
@@ -420,7 +475,7 @@ class FirstLevelTextNode(TextNode):
         return text.strip()
 
 
-class TextNodeSequence(Sequence):
+class TextNodeSequence(Sequence, TextNode):
     """A sequence of values retrieved from the element's content."""
 
     def get(self):
@@ -443,7 +498,7 @@ class TextNodeGrid(TextNodeSequence):
         if len(result) % self.columns != 0:
             raise ValueError(
                 'Number of elements must be divisible by %i. %s' %(
-                self.columns, getFileInfo(self)))
+                self.columns, getFileInfo(self.context)))
         return [result[i*self.columns:(i+1)*self.columns]
                 for i in range(len(result)/self.columns)]
 
@@ -456,28 +511,17 @@ class RawXMLContent(RMLAttribute):
 
     def __init__(self, *args, **kw):
         super(RawXMLContent, self).__init__(*args, **kw)
-        # Do it in here, since we have a recursive problem otherwise
-        from z3c.rml import special
-        self.handleElements = {'getName': special.GetName}
 
     def get(self):
-        # Replace what we can replace
-        for subElement in self.context.element.iterdescendants():
-            if subElement.tag in self.handleElements:
-                substitute = self.handleElements[subElement.tag](
-                    subElement, self.context)
-                substitute.process()
-        # Now create the text
         # ReportLab's paragraph parser does not like attributes from other
         # namespaces; sigh. So we have to improvize.
         text = etree.tounicode(self.context.element)
         text = text[text.find('>')+1:text.rfind('<')]
         return text
 
-
 class XMLContent(RawXMLContent):
     """Same as 'RawXMLContent', except that the whitespace is normalized."""
 
     def get(self):
-        result = super(XMLContent, self).get()
-        return result.strip().replace('\t', ' ')
+        text = super(XMLContent, self).get()
+        return text.strip().replace('\t', ' ')
