@@ -13,7 +13,7 @@
 ##############################################################################
 """Page Drawing Related Element Processing
 
-$Id: canvas.py 114908 2010-07-22 02:23:18Z srichter $
+$Id: canvas.py 128843 2012-12-21 06:58:39Z srichter $
 """
 __docformat__ = "reStructuredText"
 import zope.interface
@@ -59,6 +59,22 @@ class CanvasRMLDirective(directive.RMLDirective):
         getattr(canvas, self.callable)(**kwargs)
 
 
+class ISaveState(interfaces.IRMLDirectiveSignature):
+    """Saves the current canvas state."""
+
+class SaveState(CanvasRMLDirective):
+    signature = ISaveState
+    callable = 'saveState'
+
+
+class IRestoreState(interfaces.IRMLDirectiveSignature):
+    """Saves the current canvas state."""
+
+class RestoreState(CanvasRMLDirective):
+    signature = IRestoreState
+    callable = 'restoreState'
+
+
 class IDrawString(interfaces.IRMLDirectiveSignature):
     """Draws a simple string (left aligned) onto the canvas at the specified
     location."""
@@ -75,7 +91,7 @@ class IDrawString(interfaces.IRMLDirectiveSignature):
                      u'string.'),
         required=True)
 
-    text = attr.TextNode(
+    text = attr.RawXMLContent(
         title=u'Text',
         description=(u'The string/text that is put onto the canvas.'),
         required=True)
@@ -84,6 +100,31 @@ class DrawString(CanvasRMLDirective):
     signature = IDrawString
     callable = 'drawString'
 
+    def getPageNumber(self, elem, canvas):
+        return str(canvas.getPageNumber() + int(elem.get('countingFrom', 1)) - 1)
+
+    def getName(self, elem, canvas):
+        return attr.getManager(self).names[elem.get('id')]
+
+    handleElements = {'pageNumber': getPageNumber,
+                      'getName': getName}
+
+    def _getText(self, node, canvas):
+        text = node.text or u''
+        for sub in node.iterdescendants():
+            if sub.tag in self.handleElements:
+                text += self.handleElements[sub.tag](self, sub, canvas)
+            else:
+                self._getText(sub, canvas)
+        text += node.tail or u''
+        return text
+
+    def process(self):
+        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
+        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
+        kwargs['text'] = self._getText(self.element, canvas).strip()
+        getattr(canvas, self.callable)(**kwargs)
+
 class IDrawRightString(IDrawString):
     """Draws a simple string (right aligned) onto the canvas at the specified
     location."""
@@ -180,6 +221,17 @@ class IRectangle(IShape):
         description=u'The radius of the rounded corners.',
         required=False)
 
+    href = attr.Text(
+        title=u'Link URL',
+        description=u'When specified, the rectangle becomes a link to that URL.',
+        required=False)
+
+    destination = attr.Text(
+        title=u'Link Destination',
+        description=(u'When specified, the rectangle becomes a link to that '
+                     u'destination.'),
+        required=False)
+
 class Rectangle(CanvasRMLDirective):
     signature = IRectangle
     callable = 'rect'
@@ -188,8 +240,24 @@ class Rectangle(CanvasRMLDirective):
     def process(self):
         if 'round' in self.element.keys():
             self.callable = 'roundRect'
-        super(Rectangle, self).process()
+        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
+        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
+        # Create a link
+        url = kwargs.pop('href', None)
+        if url:
+            canvas.linkURL(
+                url,
+                (kwargs['x'], kwargs['y'],
+                 kwargs['x']+kwargs['width'], kwargs['y']+kwargs['height']))
+        dest = kwargs.pop('destination', None)
+        if dest:
+            canvas.linkRect(
+                '', dest,
+                (kwargs['x'], kwargs['y'],
+                 kwargs['x']+kwargs['width'], kwargs['y']+kwargs['height']))
 
+        # Render the rectangle
+        getattr(canvas, self.callable)(**kwargs)
 
 class IGrid(interfaces.IRMLDirectiveSignature):
     """A shape to be drawn on the canvas."""
@@ -490,6 +558,13 @@ class IPath(IShape):
         default=False,
         required=False)
 
+    clip = attr.Boolean(
+        title=u'Clip Path',
+        description=(u"A flag specifying whether the path should clip "
+                     u"overlapping elements."),
+        default=False,
+        required=False)
+
 class Path(CanvasRMLDirective):
     signature = IPath
     factories = {
@@ -527,7 +602,10 @@ class Path(CanvasRMLDirective):
         if kwargs.pop('close', False):
             self.path.close()
 
-        canvas.drawPath(self.path, **kwargs)
+        if kwargs.pop('clip', False):
+            canvas.clipPath(self.path, **kwargs)
+        else:
+            canvas.drawPath(self.path, **kwargs)
 
 
 class IFill(interfaces.IRMLDirectiveSignature):
@@ -582,18 +660,38 @@ class SetFont(CanvasRMLDirective):
     attrMapping = {'name': 'psfontname'}
 
 
+class ISetFontSize(interfaces.IRMLDirectiveSignature):
+    """Set the font size."""
+
+    size = attr.Measurement(
+        title=u'Size',
+        description=(u'The font size.'),
+        required=True)
+
+    leading = attr.Measurement(
+        title=u'Leading',
+        description=(u'The font leading.'),
+        required=False)
+
+class SetFontSize(CanvasRMLDirective):
+    signature = ISetFontSize
+    callable = 'setFontSize'
+
+
 class IScale(interfaces.IRMLDirectiveSignature):
     """Scale the drawing using x and y scaling factors."""
 
     sx = attr.Float(
         title=u'X-Scaling-Factor',
         description=(u'The scaling factor applied on x-coordinates.'),
-        required=True)
+        default=1,
+        required=False)
 
     sy = attr.Float(
         title=u'Y-Scaling-Factor',
         description=(u'The scaling factor applied on y-coordinates.'),
-        required=True)
+        default=1,
+        required=False)
 
 class Scale(CanvasRMLDirective):
     signature = IScale
@@ -720,10 +818,87 @@ class LineMode(CanvasRMLDirective):
             canvas.setDash(kw['dash'])
 
 
+class IBookmark(interfaces.IRMLDirectiveSignature):
+    """
+    This creates a bookmark to the current page which can be referred to with
+    the given key elsewhere. (Used inside a page drawing.)
+    """
+
+    name = attr.Text(
+        title=u'Name',
+        description=u'The name of the bookmark.',
+        required=True)
+
+    fit = attr.Choice(
+        title=u'Fit',
+        description=u'The Fit Type.',
+        choices=('XYZ', 'Fit', 'FitH', 'FitV', 'FitR'),
+        required=False)
+
+    zoom = attr.Float(
+        title=u'Zoom',
+        description=u'The zoom level when clicking on the bookmark.',
+        required=False)
+
+    x = attr.Measurement(
+        title=u'X-Position',
+        description=u'The x-position.',
+        required=False)
+
+    y = attr.Measurement(
+        title=u'Y-Position',
+        description=u'The y-position.',
+        required=False)
+
+class Bookmark(CanvasRMLDirective):
+    signature = IBookmark
+
+    def process(self):
+        args = dict(self.getAttributeValues())
+        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
+        args['left'], args['top'] = canvas.absolutePosition(args['x'], args['y'])
+        canvas.bookmarkPage(**args)
+
+
+class IPlugInGraphic(interfaces.IRMLDirectiveSignature):
+    """Inserts a custom graphic developed in Python."""
+
+    module = attr.String(
+        title=u'Module',
+        description=u'The Python module in which the flowable is located.',
+        required=True)
+
+    function = attr.String(
+        title=u'Function',
+        description=(u'The name of the factory function within the module '
+                     u'that returns the custom flowable.'),
+        required=True)
+
+    params = attr.TextNode(
+        title=u'Parameters',
+        description=(u'A list of parameters encoded as a long string.'),
+        required=False)
+
+class PlugInGraphic(CanvasRMLDirective):
+    signature = IPlugInGraphic
+
+    def process(self):
+        modulePath, functionName, params = self.getAttributeValues(
+            valuesOnly=True)
+        module = __import__(modulePath, {}, {}, [modulePath])
+        function = getattr(module, functionName)
+        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
+        function(canvas, params)
+
+
 class IDrawing(interfaces.IRMLDirectiveSignature):
     """A container directive for all directives that draw directly on the
     cnavas."""
     occurence.containing(
+        # State Manipulation
+        occurence.ZeroOrMore('saveState', ISaveState),
+        occurence.ZeroOrMore('restoreState', IRestoreState),
+        # String Drawing
         occurence.ZeroOrMore('drawString', IDrawString),
         occurence.ZeroOrMore('drawRightString', IDrawRightString),
         occurence.ZeroOrMore('drawCenteredString', IDrawCenteredString),
@@ -744,6 +919,7 @@ class IDrawing(interfaces.IRMLDirectiveSignature):
         occurence.ZeroOrMore('fill', IFill),
         occurence.ZeroOrMore('stroke', IStroke),
         occurence.ZeroOrMore('setFont', ISetFont),
+        occurence.ZeroOrMore('setFontSize', ISetFontSize),
         occurence.ZeroOrMore('scale', IScale),
         occurence.ZeroOrMore('translate', ITranslate),
         occurence.ZeroOrMore('rotate', IRotate),
@@ -763,12 +939,19 @@ class IDrawing(interfaces.IRMLDirectiveSignature):
         occurence.ZeroOrMore('pieChart', chart.IPieChart),
         occurence.ZeroOrMore('pieChart3D', chart.IPieChart3D),
         occurence.ZeroOrMore('spiderChart', chart.ISpiderChart),
+        # Misc
+        occurence.ZeroOrMore('bookmark', IBookmark),
+        occurence.ZeroOrMore('plugInGraphic', IPlugInGraphic),
         )
 
 class Drawing(directive.RMLDirective):
     signature = IDrawing
 
     factories = {
+        # State Management
+        'saveState': SaveState,
+        'restoreState': RestoreState,
+        # Drawing Strings
         'drawString': DrawString,
         'drawRightString': DrawRightString,
         'drawCenteredString': DrawCenteredString,
@@ -794,6 +977,7 @@ class Drawing(directive.RMLDirective):
         'fill': Fill,
         'stroke': Stroke,
         'setFont': SetFont,
+        'setFontSize': SetFontSize,
         'scale': Scale,
         'translate': Translate,
         'rotate': Rotate,
@@ -807,7 +991,10 @@ class Drawing(directive.RMLDirective):
         'linePlot3D': chart.LinePlot3D,
         'pieChart': chart.PieChart,
         'pieChart3D': chart.PieChart3D,
-        'spiderChart': chart.SpiderChart
+        'spiderChart': chart.SpiderChart,
+        # Misc
+        'bookmark': Bookmark,
+        'plugInGraphic': PlugInGraphic,
         }
 
 
