@@ -13,10 +13,11 @@
 ##############################################################################
 """Flowable Element Processing
 
-$Id: flowable.py 84838 2008-03-21 12:32:58Z rogerineichen $
+$Id: flowable.py 128843 2012-12-21 06:58:39Z srichter $
 """
 __docformat__ = "reStructuredText"
 import copy
+import logging
 import re
 import reportlab.lib.styles
 import reportlab.platypus
@@ -24,7 +25,8 @@ import reportlab.platypus.doctemplate
 import reportlab.platypus.flowables
 import reportlab.platypus.tables
 import zope.schema
-from reportlab.lib import styles
+from reportlab.lib import styles, pygments2xpre
+from xml.sax.saxutils import unescape
 from z3c.rml import attr, directive, interfaces, occurence
 from z3c.rml import form, platypus, special, stylesheet
 
@@ -141,8 +143,7 @@ class IMinimalParagraphBase(interfaces.IRMLDirectiveSignature):
         description=(u'The paragraph style that is applied to the paragraph. '
                      u'See the ``paraStyle`` tag for creating a paragraph '
                      u'style.'),
-        default=reportlab.lib.styles.getSampleStyleSheet()['Normal'],
-        required=True)
+        required=False)
 
     bulletText = attr.String(
         title=u'Bullet Character',
@@ -185,19 +186,44 @@ class IParagraphBase(IMinimalParagraphBase):
 class IPreformatted(IMinimalParagraphBase):
     """A preformatted text, similar to the <pre> tag in HTML."""
 
+    style = attr.Style(
+        title=u'Style',
+        description=(u'The paragraph style that is applied to the paragraph. '
+                     u'See the ``paraStyle`` tag for creating a paragraph '
+                     u'style.'),
+        default=reportlab.lib.styles.getSampleStyleSheet()['Code'],
+        required=False)
+
     text = attr.RawXMLContent(
         title=u'Text',
         description=(u'The text that will be layed out.'),
         required=True)
 
+    maxLineLength = attr.Integer(
+        title=u'Max Line Length',
+        description=(u'The maximum number of characters on one line.'),
+        required=False)
+
+    newLineChars = attr.Text(
+        title=u'New Line Characters',
+        description=u'The characters placed at the beginning of a wrapped line',
+        required=False)
+
 class Preformatted(Flowable):
     signature = IPreformatted
     klass = reportlab.platypus.Preformatted
 
-
 class IXPreformatted(IParagraphBase):
     """A preformatted text that allows paragraph markup."""
 
+    style = attr.Style(
+        title=u'Style',
+        description=(u'The paragraph style that is applied to the paragraph. '
+                     u'See the ``paraStyle`` tag for creating a paragraph '
+                     u'style.'),
+        default=reportlab.lib.styles.getSampleStyleSheet()['Normal'],
+        required=False)
+
     text = attr.RawXMLContent(
         title=u'Text',
         description=(u'The text that will be layed out.'),
@@ -208,6 +234,36 @@ class XPreformatted(Flowable):
     klass = reportlab.platypus.XPreformatted
 
 
+class ICodeSnippet(IXPreformatted):
+    """A code snippet with text highlighting."""
+
+    style = attr.Style(
+        title=u'Style',
+        description=(u'The paragraph style that is applied to the paragraph. '
+                     u'See the ``paraStyle`` tag for creating a paragraph '
+                     u'style.'),
+        required=False)
+
+    language = attr.String(
+        title=u'Language',
+        description=u'The language the code snippet is written in.',
+        required=False)
+
+class CodeSnippet(XPreformatted):
+    signature = ICodeSnippet
+
+    def process(self):
+        args = dict(self.getAttributeValues())
+        lang = args.pop('language', None)
+        args['text'] = unescape(args['text'])
+        if lang is not None:
+            args['text'] = pygments2xpre.pygments2xpre(
+                args['text'], lang.lower())
+        if 'style' not in args:
+            args['style'] = attr._getStyle(self, 'Code')
+        self.parent.flow.append(self.klass(**args))
+
+
 class IParagraph(IParagraphBase, stylesheet.IBaseParagraphStyle):
     """Lays out an entire paragraph."""
 
@@ -219,6 +275,7 @@ class IParagraph(IParagraphBase, stylesheet.IBaseParagraphStyle):
 class Paragraph(Flowable):
     signature = IParagraph
     klass = reportlab.platypus.Paragraph
+    defaultStyle = 'Normal'
 
     styleAttributes = zope.schema.getFieldNames(stylesheet.IBaseParagraphStyle)
 
@@ -232,6 +289,8 @@ class Paragraph(Flowable):
 
     def process(self):
         args = dict(self.getAttributeValues(ignore=self.styleAttributes))
+        if 'style' not in args:
+            args['style'] = attr._getStyle(self, self.defaultStyle)
         args['style'] = self.processStyle(args['style'])
         self.parent.flow.append(self.klass(**args))
 
@@ -239,61 +298,56 @@ class Paragraph(Flowable):
 class ITitle(IParagraph):
     """The title is a simple paragraph with a special title style."""
 
-    style = attr.Style(
-        title=u'Style',
-        description=(u'The paragraph style that is applied to the paragraph. '
-                     u'See the ``paraStyle`` tag for creating a paragraph '
-                     u'style.'),
-        default=reportlab.lib.styles.getSampleStyleSheet()['Title'],
-        required=True)
-
 class Title(Paragraph):
     signature = ITitle
-
+    defaultStyle = 'Title'
 
 class IHeading1(IParagraph):
     """Heading 1 is a simple paragraph with a special heading 1 style."""
 
-    style = attr.Style(
-        title=u'Style',
-        description=(u'The paragraph style that is applied to the paragraph. '
-                     u'See the ``paraStyle`` tag for creating a paragraph '
-                     u'style.'),
-        default=reportlab.lib.styles.getSampleStyleSheet()['Heading1'],
-        required=True)
-
 class Heading1(Paragraph):
     signature = IHeading1
+    defaultStyle = 'Heading1'
 
 
 class IHeading2(IParagraph):
     """Heading 2 is a simple paragraph with a special heading 2 style."""
 
-    style = attr.Style(
-        title=u'Style',
-        description=(u'The paragraph style that is applied to the paragraph. '
-                     u'See the ``paraStyle`` tag for creating a paragraph '
-                     u'style.'),
-        default=reportlab.lib.styles.getSampleStyleSheet()['Heading2'],
-        required=True)
-
 class Heading2(Paragraph):
     signature = IHeading2
+    defaultStyle = 'Heading2'
 
 
 class IHeading3(IParagraph):
     """Heading 3 is a simple paragraph with a special heading 3 style."""
 
-    style = attr.Style(
-        title=u'Style',
-        description=(u'The paragraph style that is applied to the paragraph. '
-                     u'See the ``paraStyle`` tag for creating a paragraph '
-                     u'style.'),
-        default=reportlab.lib.styles.getSampleStyleSheet()['Heading3'],
-        required=True)
-
 class Heading3(Paragraph):
     signature = IHeading3
+    defaultStyle = 'Heading3'
+
+
+class IHeading4(IParagraph):
+    """Heading 4 is a simple paragraph with a special heading 4 style."""
+
+class Heading4(Paragraph):
+    signature = IHeading4
+    defaultStyle = 'Heading4'
+
+
+class IHeading5(IParagraph):
+    """Heading 5 is a simple paragraph with a special heading 5 style."""
+
+class Heading5(Paragraph):
+    signature = IHeading5
+    defaultStyle = 'Heading5'
+
+
+class IHeading6(IParagraph):
+    """Heading 6 is a simple paragraph with a special heading 6 style."""
+
+class Heading6(Paragraph):
+    signature = IHeading6
+    defaultStyle = 'Heading6'
 
 
 class ITableCell(interfaces.IRMLDirectiveSignature):
@@ -470,6 +524,18 @@ class ITableCell(interfaces.IRMLDirectiveSignature):
         description=u'The space of the line right of the cell.',
         required=False)
 
+    href = attr.Text(
+        title=u'Link URL',
+        description=u'When specified, the cell becomes a link to that URL.',
+        required=False)
+
+    destination = attr.Text(
+        title=u'Link Destination',
+        description=(u'When specified, the cell becomes a link to that '
+                     u'destination.'),
+        required=False)
+
+
 class TableCell(directive.RMLDirective):
     signature = ITableCell
     styleAttributesMapping = (
@@ -492,6 +558,8 @@ class TableCell(directive.RMLDirective):
                         'lineLeftCap', 'lineLeftCount', 'lineLeftSpace')),
         ('LINEAFTER', ('lineRightThickness', 'lineRightColor',
                        'lineRightCap', 'lineRightCount', 'lineRightSpace')),
+        ('HREF', ('href',)),
+        ('DESTINATION', ('destination',)),
         )
 
     def processStyle(self):
@@ -755,7 +823,7 @@ class IKeepTogether(interfaces.IRMLDirectiveSignature):
         description=u'The maximum height the flowables are allotted.',
         default=None,
         required=False)
-    
+
 class KeepTogether(Flowable):
     signature = IKeepTogether
     klass = reportlab.platypus.flowables.KeepTogether
@@ -771,6 +839,50 @@ class KeepTogether(Flowable):
         frame = self.klass(flow.flow, **args)
         self.parent.flow.append(frame)
 
+class IImage(interfaces.IRMLDirectiveSignature):
+    """An image."""
+
+    src = attr.Image(
+        title=u'Image Source',
+        description=u'The file that is used to extract the image data.',
+        onlyOpen=True,
+        required=True)
+
+    width = attr.Measurement(
+        title=u'Image Width',
+        description=u'The width of the image.',
+        required=False)
+
+    height = attr.Measurement(
+        title=u'Image Height',
+        description=u'The height the image.',
+        required=False)
+
+    mask = attr.Color(
+        title=u'Mask',
+        description=u'The color mask used to render the image.',
+        required=False)
+
+    vAlign = attr.Choice(
+        title=u'Vertical Alignment',
+        description=u'The vertical alignment of the image.',
+        choices=interfaces.VALIGN_TEXT_CHOICES,
+        required=False)
+
+class Image(Flowable):
+    signature = IImage
+    klass = reportlab.platypus.flowables.Image
+    attrMapping = {'src': 'filename'}
+
+    def process(self):
+        args = dict(self.getAttributeValues(attrMapping=self.attrMapping))
+        vAlign = args.pop('vAlign', None)
+        img = self.klass(**args)
+        if vAlign:
+            img.vAlign = vAlign
+        self.parent.flow.append(img)
+
+
 class IImageAndFlowables(interfaces.IRMLDirectiveSignature):
     """An image with flowables around it."""
 
@@ -931,7 +1043,7 @@ class FixedSize(Flowable):
         self.parent.flow.append(frame)
 
 
-class IBookmark(interfaces.IRMLDirectiveSignature):
+class IBookmarkPage(interfaces.IRMLDirectiveSignature):
     """
     This creates a bookmark to the current page which can be referred to with
     the given key elsewhere.
@@ -947,10 +1059,20 @@ class IBookmark(interfaces.IRMLDirectiveSignature):
         description=u'The name of the bookmark.',
         required=True)
 
-    fitType = attr.Choice(
-        title=u'Fit Type',
+    fit = attr.Choice(
+        title=u'Fit',
         description=u'The Fit Type.',
-        choices=('Fit', 'FitH', 'FitV', 'FitR'),
+        choices=('XYZ', 'Fit', 'FitH', 'FitV', 'FitR'),
+        required=False)
+
+    top = attr.Measurement(
+        title=u'Top',
+        description=u'The top position.',
+        required=False)
+
+    bottom = attr.Measurement(
+        title=u'Bottom',
+        description=u'The bottom position.',
         required=False)
 
     left = attr.Measurement(
@@ -963,27 +1085,46 @@ class IBookmark(interfaces.IRMLDirectiveSignature):
         description=u'The right position.',
         required=False)
 
-    top = attr.Measurement(
-        title=u'Top',
-        description=u'The top position.',
-        required=False)
-
-    right = attr.Measurement(
-        title=u'Right',
-        description=u'The right position.',
-        required=False)
-
     zoom = attr.Float(
         title=u'Zoom',
         description=u'The zoom level when clicking on the bookmark.',
         required=False)
 
-class Bookmark(Flowable):
-    signature = IBookmark
+class BookmarkPage(Flowable):
+    signature = IBookmarkPage
     klass = platypus.BookmarkPage
     attrMapping = {'name': 'key', 'fitType': 'fit'}
 
 
+class IBookmark(interfaces.IRMLDirectiveSignature):
+    """
+    This creates a bookmark to the current page which can be referred to with
+    the given key elsewhere. (Used inside a story.)
+    """
+
+    name = attr.Text(
+        title=u'Name',
+        description=u'The name of the bookmark.',
+        required=True)
+
+    x = attr.Measurement(
+        title=u'X Coordinate',
+        description=u'The x-position of the bookmark.',
+        default=0,
+        required=False)
+
+    y = attr.Measurement(
+        title=u'Y Coordinate',
+        description=u'The y-position of the bookmark.',
+        default=0,
+        required=False)
+
+class Bookmark(Flowable):
+    signature = IBookmark
+    klass = platypus.Bookmark
+    attrMapping = {'name': 'key', 'x': 'relativeX', 'y': 'relativeY'}
+
+
 class ILink(interfaces.IRMLDirectiveSignature):
     """Place an internal link around a set of flowables."""
 
@@ -1118,6 +1259,170 @@ class OutlineAdd(Flowable):
     klass = platypus.OutlineAdd
 
 
+class NamedStringFlowable(reportlab.platypus.flowables.Flowable):
+
+    def __init__(self, manager, id, value):
+        reportlab.platypus.flowables.Flowable.__init__(self)
+        self.manager = manager
+        self.id = id
+        self.value = value
+
+    def wrap(self, *args):
+        return (0, 0)
+
+    def draw(self):
+        self.manager.names[self.id] = self.value
+
+
+class INamedString(interfaces.IRMLDirectiveSignature):
+    """Defines a name for a string."""
+
+    id = attr.String(
+        title=u'Id',
+        description=u'The id under which the value will be known.',
+        required=True)
+
+    value = attr.XMLContent(
+        title=u'Value',
+        description=u'The text that is displayed if the id is called.',
+        required=True)
+
+class NamedString(directive.RMLDirective):
+    signature = INamedString
+
+    def process(self):
+        id, value = self.getAttributeValues(valuesOnly=True)
+        manager = attr.getManager(self)
+        # We have to delay assigning values, otherwise the last one wins.
+        self.parent.flow.append(NamedStringFlowable(manager, id, value))
+
+
+class IShowIndex(interfaces.IRMLDirectiveSignature):
+    """Creates an index in the document."""
+
+    name = attr.String(
+        title=u'Name',
+        description=u'The name of the index.',
+        default='index',
+        required=False)
+
+    dot = attr.String(
+        title=u'Dot',
+        description=u'The character to use as a dot.',
+        required=False)
+
+    style = attr.Style(
+        title=u'Style',
+        description=u'The paragraph style that is applied to the index. ',
+        required=False)
+
+    tableStyle = attr.Style(
+        title=u'Table Style',
+        description=u'The table style that is applied to the index layout. ',
+        required=False)
+
+class ShowIndex(directive.RMLDirective):
+    signature = IShowIndex
+
+    def process(self):
+        args = dict(self.getAttributeValues())
+        manager = attr.getManager(self)
+        index = manager.indexes[args['name']]
+        args['format'] = index.formatFunc.__name__[8:]
+        args['offset'] = index.offset
+        index.setup(**args)
+        self.parent.flow.append(index)
+
+
+class IBaseLogCall(interfaces.IRMLDirectiveSignature):
+
+    message = attr.RawXMLContent(
+        title=u'Message',
+        description=u'The message to be logged.',
+        required=True)
+
+class LogCallFlowable(reportlab.platypus.flowables.Flowable):
+
+    def __init__(self, logger, level, message):
+        self.logger = logger
+        self.level = level
+        self.message = message
+
+    def wrap(self, *args):
+        return (0, 0)
+
+    def draw(self):
+        self.logger.log(self.level, self.message)
+
+class BaseLogCall(directive.RMLDirective):
+    signature = IBaseLogCall
+    level = None
+
+    def process(self):
+        message = self.getAttributeValues(
+            select=('message',), valuesOnly=True)[0]
+        manager = attr.getManager(self)
+        self.parent.flow.append(
+            LogCallFlowable(manager.logger, self.level, message))
+
+class ILog(IBaseLogCall):
+    """Log message at DEBUG level."""
+
+    level = attr.Choice(
+        title=u'Level',
+        description=u'The default log level.',
+        choices=interfaces.LOG_LEVELS,
+        doLower=False,
+        default=logging.INFO,
+        required=True)
+
+class Log(BaseLogCall):
+    signature = ILog
+
+    @property
+    def level(self):
+        return self.getAttributeValues(select=('level',), valuesOnly=True)[0]
+
+class IDebug(IBaseLogCall):
+    """Log message at DEBUG level."""
+
+class Debug(BaseLogCall):
+    signature = IDebug
+    level = logging.DEBUG
+
+
+class IInfo(IBaseLogCall):
+    """Log message at INFO level."""
+
+class Info(BaseLogCall):
+    signature = IInfo
+    level = logging.INFO
+
+
+class IWarning(IBaseLogCall):
+    """Log message at WARNING level."""
+
+class Warning(BaseLogCall):
+    signature = IWarning
+    level = logging.WARNING
+
+
+class IError(IBaseLogCall):
+    """Log message at ERROR level."""
+
+class Error(BaseLogCall):
+    signature = IError
+    level = logging.ERROR
+
+
+class ICritical(IBaseLogCall):
+    """Log message at CRITICAL level."""
+
+class Critical(BaseLogCall):
+    signature = ICritical
+    level = logging.CRITICAL
+
+
 class IFlow(interfaces.IRMLDirectiveSignature):
     """A list of flowables."""
     occurence.containing(
@@ -1125,6 +1430,7 @@ class IFlow(interfaces.IRMLDirectiveSignature):
         occurence.ZeroOrMore('illustration', IIllustration),
         occurence.ZeroOrMore('pre', IPreformatted),
         occurence.ZeroOrMore('xpre', IXPreformatted),
+        occurence.ZeroOrMore('codesnippet', ICodeSnippet),
         occurence.ZeroOrMore('plugInFlowable', IPluginFlowable),
         occurence.ZeroOrMore('barCodeFlowable', IBarCodeFlowable),
         occurence.ZeroOrMore('outlineAdd', IOutlineAdd),
@@ -1132,6 +1438,9 @@ class IFlow(interfaces.IRMLDirectiveSignature):
         occurence.ZeroOrMore('h1', IHeading1),
         occurence.ZeroOrMore('h2', IHeading2),
         occurence.ZeroOrMore('h3', IHeading3),
+        occurence.ZeroOrMore('h4', IHeading4),
+        occurence.ZeroOrMore('h5', IHeading5),
+        occurence.ZeroOrMore('h6', IHeading6),
         occurence.ZeroOrMore('para', IParagraph),
         occurence.ZeroOrMore('blockTable', IBlockTable),
         occurence.ZeroOrMore('nextFrame', INextFrame),
@@ -1141,14 +1450,24 @@ class IFlow(interfaces.IRMLDirectiveSignature):
         occurence.ZeroOrMore('condPageBreak', IConditionalPageBreak),
         occurence.ZeroOrMore('keepInFrame', IKeepInFrame),
         occurence.ZeroOrMore('keepTogether', IKeepTogether),
+        occurence.ZeroOrMore('img', IImage),
         occurence.ZeroOrMore('imageAndFlowables', IImageAndFlowables),
         occurence.ZeroOrMore('pto', IPTO),
         occurence.ZeroOrMore('indent', IIndent),
         occurence.ZeroOrMore('fixedSize', IFixedSize),
+        occurence.ZeroOrMore('bookmarkPage', IBookmarkPage),
         occurence.ZeroOrMore('bookmark', IBookmark),
         occurence.ZeroOrMore('link', ILink),
         occurence.ZeroOrMore('hr', IHorizontalRow),
+        occurence.ZeroOrMore('showIndex', IShowIndex),
         occurence.ZeroOrMore('name', special.IName),
+        occurence.ZeroOrMore('namedString', INamedString),
+        occurence.ZeroOrMore('log', ILog),
+        occurence.ZeroOrMore('debug', IDebug),
+        occurence.ZeroOrMore('info', IInfo),
+        occurence.ZeroOrMore('warning', IWarning),
+        occurence.ZeroOrMore('error', IError),
+        occurence.ZeroOrMore('critical', ICritical),
         )
 
 class Flow(directive.RMLDirective):
@@ -1159,6 +1478,7 @@ class Flow(directive.RMLDirective):
         'illustration': Illustration,
         'pre': Preformatted,
         'xpre': XPreformatted,
+        'codesnippet': CodeSnippet,
         'plugInFlowable': PluginFlowable,
         'barCodeFlowable': BarCodeFlowable,
         'outlineAdd': OutlineAdd,
@@ -1167,6 +1487,9 @@ class Flow(directive.RMLDirective):
         'h1': Heading1,
         'h2': Heading2,
         'h3': Heading3,
+        'h4': Heading4,
+        'h5': Heading5,
+        'h6': Heading6,
         'para': Paragraph,
         # Table Flowable
         'blockTable': BlockTable,
@@ -1178,15 +1501,26 @@ class Flow(directive.RMLDirective):
         'condPageBreak': ConditionalPageBreak,
         'keepInFrame': KeepInFrame,
         'keepTogether': KeepTogether,
+        'img': Image,
         'imageAndFlowables': ImageAndFlowables,
         'pto': PTO,
         'indent': Indent,
         'fixedSize': FixedSize,
+        'bookmarkPage': BookmarkPage,
         'bookmark': Bookmark,
         'link': Link,
         'hr': HorizontalRow,
+        'showIndex': ShowIndex,
         # Special Elements
         'name': special.Name,
+        'namedString': NamedString,
+        # Logging
+        'log': Log,
+        'debug': Debug,
+        'info': Info,
+        'warning': Warning,
+        'error': Error,
+        'critical': Critical,
         }
 
     def __init__(self, *args, **kw):
