@@ -13,7 +13,7 @@
 ##############################################################################
 """Style Related Element Processing
 
-$Id: platypus.py 121449 2011-04-19 16:00:38Z menesis $
+$Id: platypus.py 128833 2012-12-21 02:02:02Z srichter $
 """
 __docformat__ = "reStructuredText"
 import reportlab.platypus.flowables
@@ -79,6 +79,11 @@ class BookmarkPage(BaseFlowable):
         self.canv.bookmarkPage(*self.args, **self.kw)
 
 
+class Bookmark(BaseFlowable):
+    def draw(self):
+        self.canv.bookmarkHorizontal(*self.args, **self.kw)
+
+
 class OutlineAdd(BaseFlowable):
     def draw(self):
         if self.kw.get('key', None) is None:
