@@ -13,7 +13,7 @@
 ##############################################################################
 """Style Related Element Processing
 
-$Id: stylesheet.py 114942 2010-07-22 19:22:10Z srichter $
+$Id: stylesheet.py 128798 2012-12-20 04:07:15Z srichter $
 """
 __docformat__ = "reStructuredText"
 import copy
@@ -71,6 +71,12 @@ class IBaseParagraphStyle(interfaces.IRMLDirectiveSignature):
         description=u'The indentation of the first line in the paragraph.',
         required=False)
 
+    alignment = attr.Choice(
+        title=u'Alignment',
+        description=u'The text alignment.',
+        choices=interfaces.ALIGN_CHOICES,
+        required=False)
+
     spaceBefore = attr.Measurement(
         title=u'Space Before',
         description=u'The vertical space before the paragraph.',
@@ -81,12 +87,6 @@ class IBaseParagraphStyle(interfaces.IRMLDirectiveSignature):
         description=u'The vertical space after the paragraph.',
         required=False)
 
-    alignment = attr.Choice(
-        title=u'Alignment',
-        description=u'The text alignment.',
-        choices=interfaces.ALIGN_CHOICES,
-        required=False)
-
     bulletFontName = attr.String(
         title=u'Bullet Font Name',
         description=u'The font in which the bullet character will be rendered.',
@@ -112,33 +112,24 @@ class IBaseParagraphStyle(interfaces.IRMLDirectiveSignature):
         description=u'The background color of the paragraph.',
         required=False)
 
-    keepWithNext = attr.Boolean(
-        title=u'Keep with Next',
-        description=(u'When set, this paragraph will always be in the same '
-                     u'frame as the following flowable.'),
-        required=False)
-
     wordWrap = attr.String(
         title=u'Word Wrap Method',
         description=(u'When set to "CJK", invoke CJK word wrapping'),
         required=False)
 
-    borderColor = attr.Color(
-        title=u'Border Color',
-        description=u'The color in which the paragraph border will appear.',
-        required=False)
-
     borderWidth = attr.Measurement(
         title=u'Paragraph Border Width',
         description=u'The width of the paragraph border.',
         required=False)
 
-    borderPadding = attr.Sequence(
+    borderPadding = attr.Padding(
         title=u'Paragraph Border Padding',
         description=u'Padding of the paragraph.',
-        value_type=attr.Integer(),
-        min_length=1,
-        max_length=4,
+        required=False)
+
+    borderColor = attr.Color(
+        title=u'Border Color',
+        description=u'The color in which the paragraph border will appear.',
         required=False)
 
     borderRadius = attr.Measurement(
@@ -146,6 +137,48 @@ class IBaseParagraphStyle(interfaces.IRMLDirectiveSignature):
         description=u'The radius of the paragraph border.',
         required=False)
 
+    allowWidows = attr.Boolean(
+        title=u'Allow Widows',
+        description=(u'Allow widows.'),
+        required=False)
+
+    allowOrphans = attr.Boolean(
+        title=u'Allow Orphans',
+        description=(u'Allow orphans.'),
+        required=False)
+
+    textTransforms = attr.Choice(
+        title=u'Text Transforms',
+        description=u'Text transformations.',
+        choices=interfaces.TEXT_TRANSFORM_CHOICES,
+        required=False)
+
+    endDots = attr.String(
+        title=u'End Dots',
+        description=u'Characters/Dots at the end of a paragraph.',
+        required=False)
+
+    # Attributes not part of the official style attributes, but are accessed
+    # by the paragraph renderer.
+
+    keepWithNext = attr.Boolean(
+        title=u'Keep with Next',
+        description=(u'When set, this paragraph will always be in the same '
+                     u'frame as the following flowable.'),
+        required=False)
+
+    pageBreakBefore = attr.Boolean(
+        title=u'Page Break Before',
+        description=(u'Specifies whether a page break should be inserted '
+                     u'before the directive.'),
+        required=False)
+
+    frameBreakBefore = attr.Boolean(
+        title=u'Frame Break Before',
+        description=(u'Specifies whether a frame break should be inserted '
+                     u'before the directive.'),
+        required=False)
+
 
 class IParagraphStyle(IBaseParagraphStyle):
     """Defines a paragraph style and gives it a name."""
@@ -171,10 +204,11 @@ class ParagraphStyle(directive.RMLDirective):
 
     def process(self):
         kwargs = dict(self.getAttributeValues())
-
         parent = kwargs.pop(
             'parent', reportlab.lib.styles.getSampleStyleSheet()['Normal'])
+        name = kwargs.pop('name')
         style = copy.deepcopy(parent)
+        style.name = name[6:] if name.startswith('style.') else name
 
         for name, value in kwargs.items():
             setattr(style, name, value)
@@ -533,12 +567,104 @@ class BlockTableStyle(directive.RMLDirective):
         manager.styles[id] = self.style
 
 
+class IMinimalListStyle(interfaces.IRMLDirectiveSignature):
+
+    leftIndent = attr.Measurement(
+        title=u'Left Indentation',
+        description=u'General indentation on the left side.',
+        required=False)
+
+    rightIndent = attr.Measurement(
+        title=u'Right Indentation',
+        description=u'General indentation on the right side.',
+        required=False)
+
+    bulletColor = attr.Color(
+        title=u'Bullet Color',
+        description=u'The color in which the bullet will appear.',
+        required=False)
+
+    bulletFontName = attr.String(
+        title=u'Bullet Font Name',
+        description=u'The font in which the bullet character will be rendered.',
+        required=False)
+
+    bulletFontSize = attr.Measurement(
+        title=u'Bullet Font Size',
+        description=u'The font size of the bullet character.',
+        required=False)
+
+    bulletOffsetY = attr.Measurement(
+        title=u'Bullet Y-Offset',
+        description=u'The vertical offset of the bullet.',
+        required=False)
+
+    bulletDedent = attr.StringOrInt(
+        title=u'Bullet Dedent',
+        description=u'Either pixels of dedent or auto (default).',
+        required=False)
+
+    bulletDir = attr.Choice(
+        title=u'Bullet Layout Direction',
+        description=u'The layout direction of the bullet.',
+        choices=('ltr', 'rtl'),
+        required=False)
+
+    bulletFormat = attr.String(
+        title=u'Bullet Format',
+        description=u'A formatting expression for the bullet text.',
+        required=False)
+
+class IBaseListStyle(IMinimalListStyle):
+
+    start = attr.Combination(
+        title=u'Start Value',
+        description=u'The counter start value.',
+        value_types=(attr.Integer(),
+                     attr.Choice(choices=interfaces.UNORDERED_BULLET_VALUES)),
+        required=False)
+
+
+class IListStyle(IBaseListStyle):
+    """Defines a list style and gives it a name."""
+
+    name = attr.String(
+        title=u'Name',
+        description=u'The name of the style.',
+        required=True)
+
+    parent = attr.Style(
+        title=u'Parent',
+        description=(u'The list style that will be used as a base for '
+                     u'this one.'),
+        required=False)
+
+
+class ListStyle(directive.RMLDirective):
+    signature = IListStyle
+
+    def process(self):
+        kwargs = dict(self.getAttributeValues())
+        parent = kwargs.pop(
+            'parent', reportlab.lib.styles.ListStyle(name='List'))
+        name = kwargs.pop('name')
+        style = copy.deepcopy(parent)
+        style.name = name[6:] if name.startswith('style.') else name
+
+        for name, value in kwargs.items():
+            setattr(style, name, value)
+
+        manager = attr.getManager(self)
+        manager.styles[style.name] = style
+
+
 class IStylesheet(interfaces.IRMLDirectiveSignature):
     """A styleheet defines the styles that can be used in the document."""
     occurence.containing(
         occurence.ZeroOrOne('initialize', IInitialize),
         occurence.ZeroOrMore('paraStyle', IParagraphStyle),
         occurence.ZeroOrMore('blockTableStyle', IBlockTableStyle),
+        occurence.ZeroOrMore('listStyle', IListStyle),
         # TODO:
         #occurence.ZeroOrMore('boxStyle', IBoxStyle),
         )
@@ -550,4 +676,5 @@ class Stylesheet(directive.RMLDirective):
         'initialize': Initialize,
         'paraStyle': ParagraphStyle,
         'blockTableStyle': BlockTableStyle,
+        'listStyle': ListStyle,
         }
