@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2007 Zope Foundation and Contributors.
+# Copyright (c) 2007-2012 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -13,10 +13,14 @@
 ##############################################################################
 """Testing all XML Locale functionality.
 
-$Id: module.py 117264 2010-10-06 06:22:49Z icemac $
+$Id: module.py 128806 2012-12-20 16:34:40Z srichter $
 """
 import reportlab.platypus
 import reportlab.lib.styles
+import reportlab.graphics.widgets.markers
+from reportlab.graphics import renderPDF, shapes
+from reportlab.lib import colors
+from z3c.rml import interfaces, attr
 
 def myPreformatted(params):
     return reportlab.platypus.Preformatted('''
@@ -41,7 +45,7 @@ class LinkURL(reportlab.platypus.flowables.Flowable):
 
     def draw(self):
         self.canv.linkURL(self.link, None)
-        
+
 
 def linkURL(params):
     params = eval(params)
@@ -49,3 +53,28 @@ def linkURL(params):
         reportlab.platypus.Paragraph(
             params[0], reportlab.lib.styles.getSampleStyleSheet()['Normal']),
         LinkURL(*params))
+
+class IMarker(interfaces.IRMLDirectiveSignature):
+    x = attr.Measurement()
+    y = attr.Measurement()
+    dx = attr.Measurement()
+    dy = attr.Measurement()
+    size = attr.Measurement()
+    fillColor = attr.Color(acceptNone=True, required=False)
+    strokeColor = attr.Color(acceptNone=True, required=False)
+    strokeWidth = attr.Measurement(required=False)
+    arrowBarDx = attr.Measurement(required=False)
+    arrowHeight = attr.Measurement(required=False)
+
+def symbols(canvas, params):
+    args = eval('dict(%s)' %params)
+    name = args.pop('name')
+    n = args.pop('n')
+    for key, value in args.items():
+        field = IMarker[key].bind(canvas.manager)
+        args[key] = field.fromUnicode(value)
+    m = reportlab.graphics.widgets.markers.makeMarker(name, **args)
+    drawing = shapes.Drawing()
+    drawing.add(m)
+    for idx in range(n):
+        drawing.drawOn(canvas, idx*args['dx'], idx*args['dy'])
