@@ -13,7 +13,7 @@
 ##############################################################################
 """Testing all XML Locale functionality.
 
-$Id: test_rml.py 121449 2011-04-19 16:00:38Z menesis $
+$Id: test_rml.py 128841 2012-12-21 05:02:18Z srichter $
 """
 import os
 import subprocess
@@ -27,6 +27,8 @@ try:
 except ImportError:
     from PIL import Image
 
+LOG_FILE = os.path.join(os.path.dirname(__file__), 'render.log')
+
 def gs_command(path):
     return ('gs', '-q', '-sNOPAUSE', '-sDEVICE=png256',
             '-sOutputFile=%s[Page-%%d].png' % path[:-4],
@@ -60,6 +62,8 @@ class RMLRenderingTestCase(unittest.TestCase):
         attr.File.open = self._fileOpen
         del sys.modules['module']
         del sys.modules['mymodule']
+        if os.path.exists(LOG_FILE):
+            os.remove(LOG_FILE)
 
     def runTest(self):
         rml2pdf.go(self._inPath, self._outPath)
